

theory MonadComposition
imports MonadComposition_P2
begin

section {* The additional operation parameter of the monad composition - swap *}

consts
 swap :: "('a M1) M2 \<Rightarrow> ('a M2) M1"



constdefs
 prod :: "'a M2 M1 M2 \<Rightarrow> 'a M2 M1"
"prod \<equiv> mapM1 joinM2 \<circ> swap"
 dorp :: "'a M1 M2 M1 \<Rightarrow> 'a M2 M1"
"dorp \<equiv> joinM1 \<circ> mapM1 swap" 

lemma prod_dorp_eq : "mapM1 joinM2 \<circ> dorp = joinM1 \<circ> mapM1 prod"
apply(unfold dorp_def)
apply(subst o_assoc)
apply(subst mapM1_joinM1_comp2[THEN sym])
apply(subst o_assoc[THEN sym])
apply(subst mapM1_comp)
apply(fold prod_def)
by (rule refl)

lemma dorp_join : "joinM1 \<circ> (mapM1 dorp) = dorp \<circ> joinM1"
apply(unfold dorp_def)
apply(subst o_assoc[THEN sym])+
apply(subst mapM1_joinM1_comp2[THEN sym])
apply(subst mapM1_comp[THEN sym])
apply(subst o_assoc)+
apply(subst joinM1_mapM1_joinM1)
by (rule refl)


text {* The rules for swap *}
axioms
 swap_eta     : "swap \<circ> eta2 = mapM1 eta2"
 swap_map     : "swap \<circ> mapM2 (mapM1 f) = mapM1 (mapM2 f) \<circ> swap"
 swap_map_eta : "swap \<circ> mapM2 eta1 = eta1"
 prod_dorp    : "prod \<circ> mapM2 dorp = dorp \<circ> prod"


section {* Join, map, eta for composed monad and their properties *}

constdefs
 joinCM :: "'a M2 M1 M2 M1 \<Rightarrow> 'a M2 M1"
"joinCM \<equiv> (mapM1 joinM2) \<circ> (joinM1 \<circ> mapM1 swap)"
 mapCM :: "('a \<Rightarrow> 'b) \<Rightarrow> 'a M2 M1 \<Rightarrow> 'b M2 M1"
"mapCM \<equiv> mapM1 \<circ> mapM2" 
 eta :: "'a \<Rightarrow>  'a M2 M1" 
"eta \<equiv> eta1 \<circ> eta2"


lemma mon_eta_inj:
"eta x = eta y \<Longrightarrow> x = y"
apply(simp add: eta_def)
apply(drule mon_eta_inj1)
apply(erule mon_eta_inj2)
done

lemma mapCM_eta : " (mapCM f) \<circ> eta = eta \<circ> f " 
apply(simp add: mapCM_def eta_def)
apply(subst o_assoc)
apply(subst mapM1_eta1)
apply(subst o_assoc[THEN sym])
apply(subst mapM2_eta2)
apply(subst o_assoc)
by (rule refl)

lemma mapCM_comp : " (mapCM f) \<circ> (mapCM g) = mapCM (f \<circ> g) "
apply (simp add: mapCM_def)
apply(subst mapM1_comp)
apply(subst mapM2_comp)
by (rule refl)


lemma mapCM_id : "mapCM id = id"
apply(simp add: mapCM_def)
apply(subst mapM2_id)
by(rule mapM1_id)

lemma joinCM_eta : "joinCM \<circ> eta = id"
apply (simp add: joinCM_def eta_def)
apply(subst o_assoc[THEN sym])+
apply(subst o_assoc[where g=eta1])
apply(subst mapM1_eta1)
apply(subst o_assoc[THEN sym])+
apply(subst o_assoc[where g=eta1])
apply(subst joinM1_eta1)
apply simp
apply(subst swap_eta)
apply(subst mapM1_comp)
apply(subst joinM2_eta2)
by(rule mapM1_id)


lemma mapCM_joinCM_comp : " joinCM \<circ> (mapCM eta) = id "
apply(simp add: joinCM_def mapCM_def eta_def)
apply(subst mapM2_comp[THEN sym])
apply(subst mapM1_comp[THEN sym])
apply(subst o_assoc[THEN sym])+
apply(subst o_assoc[where f="mapM1 swap"])
apply(subst mapM1_comp)
apply(subst swap_map_eta)
apply(subst mapM1_comp)
apply(subst mapM1_eta1[THEN sym])
apply(subst mapM1_comp[THEN sym])
apply(subst o_assoc[where f="joinM1"])
apply(subst mapM1_joinM1_comp2)
apply(subst o_assoc[where g="joinM1", THEN sym])
apply(subst mapM1_joinM1_comp)
apply simp
apply(subst mapM1_comp)
apply(subst mapM2_joinM2_comp)
apply(rule mapM1_id)
done


lemma mapCM_joinCM_comp2 : " joinCM \<circ> mapCM (mapCM f) = (mapCM f) \<circ> joinCM "
apply(simp add: joinCM_def mapCM_def)
apply(subst o_assoc[THEN sym])+
apply(subst mapM1_comp)
apply(subst swap_map)
apply(subst mapM1_comp[THEN sym])
apply(subst o_assoc[where f="joinM1"])
apply(subst mapM1_joinM1_comp2)
apply(subst o_assoc)+
apply(subst mapM1_comp)
apply(subst mapM2_joinM2_comp2)
apply(subst mapM1_comp[THEN sym])
by (rule refl)



lemma joinCM_mapCM_joinCM : " joinCM \<circ> mapCM joinCM = joinCM \<circ> joinCM "
apply(simp add: joinCM_def mapCM_def)
apply(subst o_assoc[THEN sym])+
apply(fold dorp_def)
apply(subst mapM2_comp[THEN sym])+
apply(subst mapM1_comp)
apply(subst o_assoc[where f="swap"])
apply(subst swap_map)
apply(subst mapM1_comp[THEN sym])+
apply(subst o_assoc[THEN sym])+
apply(subst o_assoc[where f="joinM1"])
apply(subst mapM1_joinM1_comp2)
apply(subst o_assoc[THEN sym])+
apply(subst o_assoc[where f="mapM1 joinM2"])
apply(subst mapM1_comp)
apply(subst joinM2_mapM2_joinM2)
apply(subst o_assoc[where f="joinM1"])+
apply(fold dorp_def)
apply(subst mapM1_comp[THEN sym])+
apply(subst o_assoc[THEN sym])+
apply(subst o_assoc[where g="dorp"])+
apply(subst prod_dorp_eq)
apply(subst o_assoc[THEN sym])+
apply(subst mapM1_comp)+
apply(subst prod_dorp)
apply(subst mapM1_comp[THEN sym])+
apply(subst prod_dorp_eq)
apply(subst o_assoc[where f="joinM1"])
apply(subst dorp_join)
apply(subst o_assoc)+
by (rule refl)



section {* Constructing a proper instance of the monad class *}

thymorph mcomp : MonadType \<longrightarrow> MonadComposition
type_map: [("'a M" \<mapsto> "('a M2) M1")]

instantiate_theory MonadAxms by_thymorph mcomp
op_map: [("MonadOpEta.eta" \<mapsto> "MonadComposition.eta")]

defs 
 bind_def : "(m >>= f) \<equiv> joinCM (mapCM f m)"


lemma mon_assoc:
"(s >>= t >>= u) = (s >>= (\<lambda>x. t x >>= u))"
apply(simp add: bind_def)
apply(subgoal_tac "(\<lambda>x. joinCM (mapCM u (t x))) = joinCM \<circ> mapCM u \<circ> t")
apply(erule ssubst)
apply(subst mapCM_comp[THEN sym])+
apply simp
apply(subst joinCM_mapCM_joinCM[THEN fun_cong, simplified comp_def])
apply(subst mapCM_joinCM_comp2[THEN fun_cong, simplified comp_def])
apply(rule refl)
by(rule ext, simp)


lemma mon_lunit:
"(eta x >>= t) = t x"
apply(simp add: bind_def)
apply(subst mapCM_eta[THEN fun_cong, simplified comp_def])
apply(subst joinCM_eta[THEN fun_cong, simplified comp_def])
by simp


lemma mon_runit:
"(t >>= eta) = t"
apply(simp add: bind_def)
apply(subst mapCM_joinCM_comp[THEN fun_cong, simplified comp_def])
by simp


thymorph monad_comp by_sigmorph mcomp'

end