(* 
 * For the AWE extensions adapted file: 

    Title:      Pure/Proof/proofchecker.ML
    Author:     Stefan Berghofer, TU Muenchen


 * Here proof terms are translated by given mappings during thm-derivation
 * $Id: derive_thm.sml 1162 2009-04-21 09:41:10Z maxim $
 *)


signature DERIVE_THM =
sig

  datatype AxmImg = Axm of Symtab.key
                  | Thm of Symtab.key

  val derive_thm : (Term.typ -> Term.typ) -> (Term.term -> Term.term) ->
                   Symtab.key Symtab.table -> AxmImg Symtab.table -> Symtab.key Symtab.table -> 
                   Context.theory -> Context.theory -> Proofterm.proof -> thm

  val derive_thmI : Context.theory -> Proofterm.proof -> thm

  exception DERIVE_THM of string

end

structure Derive_Thm : DERIVE_THM =
struct




exception DERIVE_THM of string

datatype AxmImg = Axm of Symtab.key
                | Thm of Symtab.key

fun prt_verb s = (if false then writeln s else ())

open Proofterm Aux;

(***** construct a theorem out of a proof term *****)


fun alpha_thm_forall_elim thy proof cterm thm =
 let val prop = prop_of thm
     val bvar_type = (case prop of
                        Const ("all", Type ("fun", [Type ("fun", [qary, _]), _])) $ A => qary
                       | _ =>  (let val _ = writeln ("forall_elim: not quantified")
                                in
                                  raise DERIVE_THM ("forall_elim: not quantified")
                                end))
     val {T=tt_type, ...} = Thm.rep_cterm cterm

     fun verb u = writeln ("forall_elim alpha-conversion failure: "^
                          ("\nleft: " ^ ((Display.string_of_ctyp o (Thm.ctyp_of thy)) bvar_type) ^
                           "\nright: "^ ((Display.string_of_ctyp o (Thm.ctyp_of thy)) tt_type)))
                    
     val aprf = (case (Alpha.typ_aconv_map (Sign.tsig_of thy) (bvar_type, tt_type)) of
                  NONE => NONE
                | SOME mp => 
                  (let val proof' = Proofterm.map_proof_terms (Alpha.map_aconv_term mp)
                                                              (Alpha.map_aconv_typ mp) proof
                   in
                     SOME proof'
                   end))
 in
  aprf
 end           



fun gen_insts alg f vs ts =
 let fun f_var l v [] = LNil
       | f_var l (v as TVar (_, s)) (t::xs) = 
         (if (length xs < l) then LNil 
          else (if (Sorts.of_sort alg (t, s)) 
                then LCons(((v, t), xs), fn () => (f_var l v xs))
                else f_var l v xs)) 
     fun inst [] rts = LCons([], fn () => LNil)
       | inst (v::xs) rts = inst_ xs (f_var (length xs) v rts)
     and inst_ xs LNil = LNil
       | inst_ xs (LCons(((v, t), rts), lys)) =
         (let val i = inst xs rts
          in
            lapp (lmap (fn x => (f v, f t)::x) i) (fn () => inst_ xs (force lys))
          end) 
 in
   inst vs ts
 end  

fun gen_insts' alg f vs ts =
 let fun f_var v lts [] = LNil
       | f_var (v as TVar (_, s)) lts (t::xs) = 
         (if (Sorts.of_sort alg (t, s))
          then LCons(((v, t), lts, xs), fn () => (f_var v (lts@[t]) xs))
          else f_var v (lts@[t]) xs) 
     fun inst [] ts = LCons([], fn () => LNil)
       | inst (v::xs) ts = inst_ xs (f_var v [] ts)
     and inst_ xs LNil = LNil
       | inst_ xs (LCons(((v, t), lts, rts), lys)) =
         (let val i = inst xs (lts@rts)
          in
             lapp (lmap (fn x => (f v, f t)::x) i) (fn () => inst_ xs (force lys))
          end)
 in
   inst vs ts
 end  






fun thms_of_proof insts_gen map_type map_term defMap axmMap thmMap src trg prf =
let fun get_prf_names p = Proofterm.fold_proof_terms
          (fold_aterms (fn Free (x, _) => Name.declare x | _ => I)) (K I) p Name.context
    val prf_names = get_prf_names prf
    val lookup = Aux.get_thm trg 


    fun thms_of_atom thy thm Ts =
    let val tvars = OldTerm.term_tvars (Thm.full_prop_of thm)
        val (fmap, thm') = Thm.varifyT' [] thm
        val dom_vars = (map TVar tvars @ (map (fn ((_, S), ixn) => TVar (ixn, S)) fmap))
        val insts = insts_gen (Aux.sort_alg thy) (Thm.ctyp_of thy) dom_vars Ts
    in
       lmap (fn i => Thm.instantiate (i, []) (forall_intr_vars (forall_intr_frees thm'))) insts
    end




    fun thms_of vs Hs (prf as (PThm (_, ((name, prop', SOME Ts), f)))) =
   (let val prf' = Proofterm.join_proof f
        val img =
         (case (Proofterm.strip_combt prf') of
           (PAxm (name', _, _), _) => (case (Symtab.lookup axmMap name') of
                                         SOME (Axm i) => Drule.implies_intr_hyps (lookup i) 
                                       | SOME (Thm i) => Drule.implies_intr_hyps (lookup i)  
                                       | NONE => (case (Symtab.lookup defMap name') of
                                                  SOME i => Drule.implies_intr_hyps (lookup i)
                                                | NONE => Drule.implies_intr_hyps (lookup name')))
         | _ => (case (Symtab.lookup thmMap name) of
                  SOME i => Drule.implies_intr_hyps (lookup i)
                | NONE => Drule.implies_intr_hyps (lookup name)
                             handle e => (let val _ = writeln ("### Theorem " ^ name ^ 
                                                               " not found in the target...")
                                              val p = (Aux.strip_thm (Aux.freeze
                                                      (Aux.reconstruct_prf' src prop' name prf')))
                                          in
                                           case (thms_of ([], get_prf_names p) [] p) of
                                            LNil => raise DERIVE_THM ("No theorems could be derived.")
                                          | LCons (t, lts) =>
                                              (let val t' = (Aux.beta_eta_convert t)
                                                   val t'' = (Drule.standard o Thm.varifyT) t'
                                               in
                                                 Drule.implies_intr_hyps t''
                                               end)
                                          end)))
     
         val TTs = map map_type Ts
      in
        thms_of_atom trg img TTs
      end)

      | thms_of _ _ (PAxm (name, _, SOME Ts)) =
        let val img = (case (Symtab.lookup axmMap name) of
                                       SOME (Axm i) => Aux.get_axiom trg i
                                     | SOME (Thm i) => Drule.implies_intr_hyps (lookup i) 
                                     | NONE => (case (Symtab.lookup defMap name) of
                                                  SOME i => Drule.implies_intr_hyps (lookup i)
                                                | NONE => Aux.get_axiom trg name)) 
            val TTs = map map_type Ts
        in
          thms_of_atom trg img TTs
        end

      | thms_of _ Hs (PBound i) = LCons(List.nth (Hs, i), fn () => LNil)
                                   
      | thms_of (vs, names) Hs (Abst (s, SOME T, prf)) =
        let val TT = map_type T
            val ([x], names') = Name.variants [s] names;
            val thms = thms_of ((x, TT) :: vs, names') Hs prf
            val cterm = (Thm.cterm_of trg (Free (x, TT)))
            fun forall_intrs LNil = LNil
              | forall_intrs (LCons(t, lts)) = 
                (case (try (Thm.forall_intr cterm) t) of
                  NONE => forall_intrs (force lts)
                | SOME t' => LCons(t', fn () => (forall_intrs (force lts)))) 
        in
           forall_intrs thms
        end

      | thms_of (vs, names) Hs (prf % SOME t) =
      let val tt' = map_term t
          val tt = Term.subst_bounds ((map Free vs), tt')
      in 
        case (try (Thm.cterm_of trg) tt) of
         NONE => LNil
       | SOME ct =>
          let val thms = thms_of (vs, names) Hs prf
              fun forall_elims_ LNil = LNil
                | forall_elims_ (LCons(at, lts)) =
                  (case (try (Thm.forall_elim ct) at) of
                    NONE => forall_elims_ (force lts)
                  | SOME at' => LCons(at', fn () => (forall_elims_ (force lts)))) 

              fun forall_elims LNil = LNil
                | forall_elims (LCons(t, lts)) =
              (case (try (Thm.forall_elim ct) t) of
                NONE => 
               (let val aprf_opt = alpha_thm_forall_elim trg prf ct t
                in
                  case aprf_opt of
                   NONE => forall_elims (force lts)
                 | SOME aprf => (let val athms = thms_of (vs, names) Hs aprf
                                 in
                                  lapp (forall_elims_ athms) (fn () => (forall_elims (force lts)))         
                                 end)
                end)
              | SOME t' => LCons(t', fn () => (forall_elims (force lts))))
                   
          in 
           forall_elims thms
          end
       end

      | thms_of (vs, names) Hs (AbsP (s, SOME t, prf)) =
      let val tt' = map_term t
          val tt = Term.subst_bounds ((map Free vs), tt')
      in
       case (try (Thm.cterm_of trg) tt) of
        NONE => LNil
      | SOME ct =>
          let val thms = thms_of (vs, names) (Thm.assume ct :: Hs) prf
              fun implies_intrs LNil = LNil
                | implies_intrs (LCons(t, lts)) = 
                (case (try (Thm.implies_intr ct) t) of
                  NONE => implies_intrs (force lts)
                | SOME t' => LCons(t', fn () => (implies_intrs (force lts)))) 
          in
            implies_intrs thms
          end
      end

      | thms_of vars Hs (prf %% prf') =
          let val thmsAB = thms_of vars Hs prf 
              val thmsA = thms_of vars Hs prf'
              val p_thms = lprod Aux.beta_eta_convert thmsAB thmsA
              fun implies_elims LNil = LNil
                | implies_elims (LCons ((tAB, tA), lts)) =
                  (case (try (Thm.implies_elim tAB) tA) of
                    NONE => implies_elims (force lts)
                  | SOME t' => LCons(t', fn () => (implies_elims (force lts))))
          in
           implies_elims p_thms
          end

  
      | thms_of _ _ (Hyp t) = 
         let val tt = map_term t
         in
          case (try (Thm.cterm_of trg) tt) of
           NONE => LNil
         | SOME ct => LCons(Thm.assume ct, fn () => LNil)
         end

      | thms_of _ _ _ = (let val _ = writeln "partial proof term"
                         in
                             raise DERIVE_THM "partial proof term"
                         end)

  in 
   (thms_of ([], prf_names) [] prf) 
  end


fun derive_thm map_type map_term defMap axmMap thmMap src trg prf =
 let val thms = thms_of_proof gen_insts map_type map_term defMap axmMap thmMap src trg (Aux.freeze prf)
 in
  case thms of
   LNil => 
   (let val _ = writeln ("### Second attempt ...")
        val thms = thms_of_proof gen_insts' map_type map_term defMap axmMap thmMap src trg (Aux.freeze prf)
    in
     case thms of
          LNil => (let val _ = writeln "No theorem could be derived"
                   in
                    raise DERIVE_THM "No theorem could be derived"
                   end)
        | LCons(t, lts) => (Drule.standard o Thm.varifyT o Aux.beta_eta_convert) t
    end)
  | LCons(t, lts) => (Drule.standard o Thm.varifyT o Aux.beta_eta_convert) t
 end




(* derive without mapping *)

fun thm_of_proofI thy prf =
  let fun get_prf_names p = Proofterm.fold_proof_terms
          (fold_aterms (fn Free (x, _) => Name.declare x | _ => I)) (K I) p Name.context
      val prf_names = get_prf_names prf
      val lookup = Aux.get_thm thy 

      fun thm_of_atom thm Ts =
       let val tvars = OldTerm.term_tvars (Thm.full_prop_of thm);
           val (fmap, thm') = Thm.varifyT' [] thm;
           val ctye = map (pairself (Thm.ctyp_of thy))
          (map TVar tvars @ map (fn ((_, S), ixn) => TVar (ixn, S)) fmap ~~ Ts)
       in
         Thm.instantiate (ctye, []) (forall_intr_vars (forall_intr_frees thm'))
       end


    fun thm_of vs Hs (prf as (PThm (_, ((name, prop', SOME Ts), _)))) =
        (let val img = Drule.implies_intr_hyps (lookup name)
         in
          thm_of_atom img Ts
         end)

      | thm_of _ _ (PAxm (name, _, SOME Ts)) =
        let val img = Aux.get_axiom thy name 
        in
          thm_of_atom img Ts
        end

      | thm_of _ Hs (PBound i) = List.nth (Hs, i)
                                   
      | thm_of (vs, names) Hs (Abst (s, SOME T, prf)) =
        let val ([x], names') = Name.variants [s] names;
            val thm = thm_of ((x, T) :: vs, names') Hs prf
        in
            Thm.forall_intr (Thm.cterm_of thy (Free (x, T))) thm
        end

      | thm_of (vs, names) Hs (prf % SOME t) =
          let val tt = Term.subst_bounds ((map Free vs), t)
              val ct = Thm.cterm_of thy tt
              val thm = thm_of (vs, names) Hs prf

              fun alpha_thm vars Hyps proof cterm thm =
               let val _ = Aux.writeln "Trying alpha conversion"
                   val prop = prop_of thm
                   val bvar_type = (case prop of
                                     Const ("all", Type ("fun", [Type ("fun", [qary, _]), _])) $ A => qary
                                   | _ =>  raise DERIVE_THM ("forall_elim: not quantified"))
                   val {T=tt_type, ...} = Thm.rep_cterm cterm

                   val athm = (case (Alpha.typ_aconv_map (Sign.tsig_of thy) (bvar_type, tt_type)) of
                                NONE => (let val _ = Output.error_msg ("forall_elim alpha-conversion failure: "^
                                        ("\nleft: " ^ ((Display.string_of_ctyp o (Thm.ctyp_of thy)) bvar_type) ^
                                         "\nright: "^ ((Display.string_of_ctyp o (Thm.ctyp_of thy)) tt_type)))
                                         in
                                           raise DERIVE_THM ("forall_elim: type mismatch")  
                                         end) 
                             | SOME mp => 
                              (let val proof' = Proofterm.map_proof_terms (Alpha.map_aconv_term mp)
                                                                          (Alpha.map_aconv_typ mp) proof
                                   val vars' = (map (fn (v, t) => (v, Alpha.map_aconv_typ mp t)) (fst vars),
                                                snd vars)
                               in
                                 thm_of vars Hyps proof'
                               end))
                in
                 athm
                end                              
               
          in 
           Thm.forall_elim ct thm
           handle THM (_, 0, _) => (let val athm = alpha_thm (vs, names) Hs prf ct thm
                                    in
                                      Thm.forall_elim ct athm
                                    end)
          end

      | thm_of (vs, names) Hs (AbsP (s, SOME t, prf)) =
          let val tt = Term.subst_bounds ((map Free vs), t)
              val ct = Thm.cterm_of thy tt
              val thm = thm_of (vs, names) (Thm.assume ct :: Hs) prf
          in
            Thm.implies_intr ct thm
          end

      | thm_of vars HsT (prf %% prf') =
          let val thmAB = Aux.beta_eta_convert (thm_of vars HsT prf)
              val thmA = Aux.beta_eta_convert (thm_of vars HsT prf')
          in
            Thm.implies_elim thmAB thmA
          end
  
      | thm_of _ _ (Hyp t) = Thm.assume (Thm.cterm_of thy t)

      | thm_of _ _ _ = raise DERIVE_THM "partial proof term"

  in 
   Aux.beta_eta_convert (thm_of ([], prf_names) [] prf) 
  end


fun derive_thmI thy prf =
 let val thm = thm_of_proofI thy (Aux.freeze prf) 
 in
  (Drule.standard o Thm.varifyT) thm
 end

end
