(* 
 * Simple HOL proof search with depth_tac and simplifier
 * $Id: prove_hol.sml 1162 2009-04-21 09:41:10Z maxim $
 *)

signature PROVE =
sig
 
 val verbose : bool ref
 val proof_search : theory -> string list -> (string * string) list -> 
                   (Symtab.key * ((string * Term.term) * Dependency.tdeps)) list -> string -> Term.term ->
                    Thm.thm option

end

structure Prove : PROVE =
struct

val verbose = ref false

fun unfold ctxt def = LocalDefs.unfold_tac ctxt [def]


infix 7 >>>=
fun (opt >>>= f) = case opt of
                       NONE => f ()
                     | SOME x => SOME x

fun SIMPSET tacf st = tacf (simpset_of (Thm.theory_of_thm st)) st;
fun CLASET tacf st = tacf (claset_of (Thm.theory_of_thm st)) st;


fun prep_defs gthm op_basis opMap cs [] = []
  | prep_defs gthm op_basis opMap cs ((x as (defnm, ((id, t), d)))::xs) =
  (if (id mem_string cs) 
   then (if (Aux.internal_id id) then prep_defs gthm op_basis opMap cs xs 
         else 
          (case (Aux.assoc_string2 (opMap, id)) of
            NONE => 
             (case (try gthm defnm) of
               NONE => prep_defs gthm op_basis opMap cs xs
             | SOME ax => (Drule.abs_def ax)::(prep_defs gthm op_basis opMap ((Aux.term_consts t)@cs) xs)) 
          | SOME b => 
            (if (b mem_string op_basis) then
             (case (try gthm defnm) of
               NONE => prep_defs gthm op_basis opMap cs xs
             | SOME ax => (Drule.abs_def ax)::(prep_defs gthm op_basis opMap ((Aux.term_consts t)@cs) xs))
            else prep_defs gthm op_basis opMap cs xs)))
   else prep_defs gthm op_basis opMap cs xs)


fun solve_tactic thy tac premsT conclT unf_defs =
 let fun solve_tac ctxt simps [] = tac simps
       | solve_tac ctxt simps (d::ds) =  
     (       (SOLVE (tac simps)) 
      ORELSE (TRY (unfold ctxt d) 
              THEN solve_tac ctxt simps ds))  
 in
  (try (Goal.prove_global thy [] premsT conclT)) 
        (fn (x as {prems, context}) => (     FIRSTGOAL (Method.insert_tac prems) 
                                        THEN solve_tac context (local_simpset_of context) unf_defs))
 end



fun proof_search thy op_basis opMap defs_ nm prop = 
 let val consts = Aux.term_consts prop
     val defs = prep_defs (Aux.get_thm thy) op_basis opMap consts (rev defs_)
     val prems = Logic.strip_imp_prems prop
     val concl = Logic.strip_imp_concl prop
 in
  (try (Goal.prove_global thy [] prems concl)) 
    (fn (x as {prems, context}) => Class_Target.default_intro_tac context prems)
>>>=
  (fn () => (let val _ = writeln ("Proof search for " ^ nm ^ " without simplifier ...")
                 val tac = (fn s => (CLASET (fn cs => depth_tac cs (!AWE.proof_search_depth) 1))) 
             in
              solve_tactic thy tac prems concl defs
             end)) 
>>>=
  (fn () => (if !AWE.proof_search_with_simp
             then (let val _ = writeln ("Proof search for " ^ nm ^ " with simplifier ...")
                       val tac' = (SIMPSET (fn simps => safe_asm_full_simp_tac simps 1)) 
                       val tac = (fn simps => safe_asm_full_simp_tac simps 1)
                   in
                     solve_tactic thy tac prems concl defs  
                   end)
             else NONE))
 end



end
 
