
(*
 * Mapping and inserting of domain elements to the target
 * $Id: transfer.sml 1172 2009-04-30 09:27:18Z maxim $
 *)

signature TRANSFER =
sig
  val transfer_class : (Term.class -> Term.sort) ->  
                       ((Term.class -> Term.sort) -> Term.term -> Term.term) -> (theory * string) -> 
                       (string * mixfix option) Symtab.table -> 
                       (string -> string) -> theory -> (string * (string * string) * theory) 
  val transfer_tycon : (Term.class -> Term.sort) -> (theory * string) -> 
                       (string * mixfix option) Symtab.table -> 
                       (string -> string) -> theory -> 
                       (string * int * (string * string) option * theory)
  val transfer_const : bool -> (theory * string) -> (Term.typ -> Term.typ) -> 
                       (string * mixfix option) Symtab.table -> (string -> string) -> theory -> 
                       (string * theory)
  val transfer_def   : bool -> (theory * string) -> string -> (Term.term -> Term.term) -> 
                       (string * mixfix option) Symtab.table -> (string -> string) -> theory -> 
                       (string * theory)
  val transfer_axiom : (theory * string) -> (Term.term -> Term.term) -> 
                       (string * mixfix option) Symtab.table -> (string -> string) -> theory -> 
                       (string * theory) 
  val transfer_syntax : (Term.typ -> Term.typ) -> string list -> theory -> theory -> theory

  val mk_arity_type_id : string -> string

end

structure Transfer : TRANSFER =
struct


fun transfer_class cmap tmap (src, cls) renames find_name trg =
let val src_alg = Aux.sort_alg src
    val trg_alg = Aux.sort_alg trg

    val sup = Sorts.super_classes src_alg cls
    val sup' = ((Sorts.minimize_sort trg_alg) o flat o (map cmap)) sup 

    val src_id = Aux.name_of_thy src
    val trg_id = Aux.name_of_thy trg
    val (q, nm) = Aux.qname_split cls
    val cls' = (case (Symtab.lookup renames cls) of
                  SOME (r, _) => (trg_id ^ "." ^ r)
                | NONE => (let val b_nm = find_name nm
                           in 
                            (trg_id ^ "." ^ b_nm)
                           end))
    val cls'' = Aux.base cls'
    val bind_cls'' = Binding.name cls''

    val opmap_suf = Logic.const_of_class        
    fun get_def x = SOME (Aux.get_axiom src ((opmap_suf x)^"_def"))
                    handle e => NONE
    fun get_info x = SOME (AxClass.get_info src x)
                     handle e => NONE
    fun prep_axs c [] = []
      | prep_axs c (ax::axs) = (let val (q, x) = (Aux.qname_split (Thm.get_name ax))
                                    val nm = (x ^ "_" ^ (string_of_int c))
                                    val attr = Aux.get_attribs ax
                                    val term = tmap (fn x => (if (x = cls) then sup' else (cmap x))) (* !!! *) 
                                               (Aux.unvarify (prop_of ax))
                                    val s_term = Syntax.string_of_term_global trg term
                                    val _ = Aux.info ("with axiom " ^ x ^ " : " ^ s_term)
                                in
                                 ((Binding.name x, attr), [term])::(prep_axs (c + 1) axs)
                                end)
                     

in
 case (get_info cls) of
  NONE =>  (let val _ = Aux.warning ("... declaring type class: " ^ cls'' ^ 
                       (case sup' of
                         [] =>  "" 
                       | [x] => (" with " ^ x ^ " as the super class")
                       | _ => (" with " ^(commas sup') ^ " as the super classes")))
            in
             (cls', (opmap_suf cls, opmap_suf cls'), trg |> AxClass.axiomatize_class_cmd (bind_cls'', sup'))
            end)
| SOME info => (let val _ = Aux.warning ("... defining axiomatic type class: " ^ cls'' ^ 
                           (case sup' of
                             [] =>  "" 
                           | [x] => (" with " ^ x ^ " as the super class")
                           | _ => (" with " ^(commas sup') ^ " as the super classes")))
                in
                  (cls', (opmap_suf cls, opmap_suf cls'), 
                  trg |> AxClass.define_class (bind_cls'', sup') [] (prep_axs 0 (#axioms info)) |> snd)
                end)
end


fun mk_arity_type_id type_id = (case (Aux.unpack type_id) of
                                         [nm] => ("arity_type_" ^ nm)
                                       | [q, nm] => (q ^ ".arity_type_" ^ nm)
                                       | [q, q', nm] => (q ^ "." ^ q' ^ ".arity_type_" ^ nm)
                                       | _ => error ("Unexpected path in type constructor name: "^type_id^
                                                     " in mk_arity_type_id"))


fun transfer_tycon classmap (src, tc) renames find_name trg =
 let val trg_id = Aux.name_of_thy trg
     val trg_alg = Aux.sort_alg trg
     val sortmap = (Sorts.minimize_sort trg_alg) o flat o (map classmap)   

     val (tc', mxfx) = (case (Symtab.lookup renames tc) of
                         SOME (r, mx) => (let val name = (trg_id ^ "." ^ r)
                                          in
                                          (case mx of
                                            SOME m => (let val _ = Syn_mng.add_type_mixfix (name, m)
                                                       in
                                                        (name, m)
                                                       end)
                                          | NONE => (name, NoSyn))
                                          end)
                       | NONE => (let val (q, nm) = Aux.qname_split tc
                                      val b_nm = find_name nm
                                      val tc' = (trg_id ^ "." ^ b_nm)
                                      val mxfx = (case (Syn_mng.get_type_mixfix tc) of
                                                   NONE => NoSyn
                                                 | SOME m => (let val _ = Syn_mng.add_type_mixfix (tc', m)
                                                              in
                                                                m
                                                              end))
                                  in
                                  (tc', mxfx)
                                  end))  
  in
    case (Aux.type_arity(src, tc)) of
     SOME (arities, i) => 
       (let val arities' = map (fn (cls, srtl) => (tc', map sortmap srtl, classmap cls)) arities
            val _ = Aux.warning ("... adding type constructor " ^ tc')
            val trg' = trg |> Sign.add_types [(Binding.name(Aux.base tc'), i, mxfx)] 
                           |> (fold AxClass.axiomatize_arity) arities'                   
        in
         (tc', i, SOME (mk_arity_type_id tc, mk_arity_type_id tc'), trg')
        end)
   | NONE => (let val _ = Aux.info ("... adding type constructor " ^ tc' ^ " (not axiomatized)")
                  val trg' = trg |> Sign.add_types [(Binding.name(Aux.base tc'), 0, mxfx)]
              in 
               (tc', 0, NONE, trg')
              end) 
  end



fun transfer_const logical (src, opnm) map_type renames find_name trg =
  let val ty = case (Aux.const_type src opnm) of
                NONE => error ("Unexpected error in transfer_const") 
              | SOME t => Aux.unvarifyT t
      val trg_id = Aux.name_of_thy trg
      fun string_of_typ thy = (Pretty.string_of o (Syntax.pretty_typ_global thy))  

      val (((path, opnm_b), opnm_f), mxfx) = 
                         (case (Symtab.lookup renames opnm) of
                            SOME (r, mx) => (let val (p, b) = Aux.qname_split r
                                                 val name = if p = "" then ((p, b), trg_id ^ "." ^ b)
                                                            else ((p, b), trg_id ^ "." ^ p ^"_"^ b)
                                             in
                                              (case mx of
                                                SOME m => (let val _ = Syn_mng.add_mixfix (snd name, m)
                                                           in
                                                            (name, m)
                                                           end)
                                              | NONE => (name, NoSyn))
                                              end)
                           | NONE => (let val (q, nm) = Aux.qname_split2 opnm
                                          val b_opnm = find_name nm
                                          val (p, b) = Aux.qname_split b_opnm
                                          val opnm' = if p = "" then ((p, b), trg_id ^ "." ^ b)
                                                      else ((p, b), trg_id ^ "." ^ p ^"_"^ b)
                                          val mxfx = (case (Syn_mng.get_mixfix opnm) of
                                                       NONE => NoSyn
                                                     | SOME m => (let val _ = Syn_mng.add_mixfix (snd opnm', m)
                                                                  in
                                                                   m
                                                                  end))
                                      in
                                        (opnm', mxfx)
                                      end))

        val ty' = map_type ty
        val _ = if logical then Aux.warning ("... adding logical constant \n" ^ opnm_f ^ " :: \"" ^ 
                               (string_of_typ trg ty') ^ "\"   " ^ (Syn_mng.string_of_mixfix mxfx))
                else (case (try (unprefix "sko_") (Aux.base opnm)) of
                       NONE => Aux.info ("... adding derived constant \n" ^ opnm_f ^ " :: \"" ^ 
                               (string_of_typ trg ty') ^ "\"   " ^ (Syn_mng.string_of_mixfix mxfx))
                     | _ => writeln ("... adding derived constant \n" ^ opnm_f ^ " :: \"" ^ 
                               (string_of_typ trg ty') ^ "\"   " ^ (Syn_mng.string_of_mixfix mxfx)))

        val trg' = (if (path = "") then trg |> Sign.add_consts_i [(Binding.name opnm_b, ty', mxfx)]
                    else trg |> Sign.add_consts_i [(Binding.name(path^"_"^opnm_b), ty', mxfx)])
  in 
   (opnm_f, trg')
  end


fun transfer_def overloaded (src, defnm) opnm map_term renames find_name trg =
 let val ax = Aux.get_axiom src defnm
     val term = (Aux.unvarify o prop_of) ax
     val trg_id = Aux.name_of_thy trg                                     
     val ((path, defnm_b), defnm_f) = 
             (case (Symtab.lookup renames defnm) of
                    SOME (r, _) => let val (p, b) = Aux.qname_split r
                                   in
                                    if p = "" then ((p, b), trg_id ^ "." ^ b)
                                    else ((p, b), trg_id ^ "." ^ p ^"_"^ b)
                                   end
                  | NONE => (let val (dq, dnm) = Aux.qname_split2 defnm
                                 val b_defnm = find_name dnm
                                 val (p, b) = Aux.qname_split b_defnm
                             in 
                               if p = "" then ((p, b), trg_id ^ "." ^ b)
                               else ((p, b), trg_id ^ "." ^ p ^"_"^ b)
                             end))
     val _ = (if overloaded 
              then Aux.info ("... adding definition (overloaded) " ^ defnm_f 
                                ^ " of the constant " ^ opnm)
              else (case (try (unprefix "sko_") (Aux.base defnm)) of
                     NONE => Aux.info ("... adding definition " ^ defnm_f
                                       ^ " of the constant " ^ opnm)
                   | _ => writeln ("... adding definition " ^ defnm_f
                                       ^ " of the constant " ^ opnm)))

     val tterm = map_term term
     val (_, trg') = 
     (if (path = "") then trg |> PureThy.add_defs overloaded [Thm.no_attributes (Binding.name defnm_b, tterm)]
      else trg |> PureThy.add_defs overloaded [Thm.no_attributes(Binding.name (path^"_"^defnm_b), tterm)])
 in
  (defnm_f, trg')
 end

fun transfer_axiom (src, axnm) map_term renames find_name trg =
 let val ax = Aux.get_axiom src axnm
     val term = (Aux.unvarify o prop_of) ax
     val trg_id = Aux.name_of_thy trg
     val ((path, axnm_b), axnm_f) = 
             (case (Symtab.lookup renames axnm) of
                    SOME (r, _) => (let val (p, b) = Aux.qname_split r
                                    in
                                     if p = "" then ((p, b), trg_id ^ "." ^ b)
                                     else ((p, b), trg_id ^ "." ^ p ^"."^ b)
                                    end)
                  | NONE => (let val (aq, anm) = Aux.qname_split2 axnm
                                 val b_axnm = find_name anm
                                 val (p, b) = Aux.qname_split b_axnm
                             in 
                               if p = "" then ((p, b), trg_id ^ "." ^ b)
                               else ((p, b), trg_id ^ "." ^ p ^"."^ b)
                             end))                          
     val _ = Aux.info ("... translating axiom " ^ axnm ^ " as " ^ axnm_f ^ " ...")
     val tterm = map_term term
     val tterm_rep = Syntax.string_of_term_global trg (Aux.varify tterm)
     val (_, trg') = 
     (if (path = "") then trg |> PureThy.add_axioms [Thm.no_attributes(Binding.name axnm_b, tterm)]
      else trg
           |> Sign.add_path path
           |> PureThy.add_axioms [Thm.no_attributes(Binding.name axnm_b, tterm)]
           ||> Sign.parent_path) 
     val _ = Aux.warning ("... added axiom " ^ axnm_f ^ "\n" ^ quote(tterm_rep))          
 in
  (axnm_f, trg')
 end




(* *** Syntax *** *)


fun transfer_syntax map_type domain source target =
  let val trgnm = Aux.name_of_thy target

      fun nts u = Aux.foldl (op @) ([], map Syn_mng.get_nonterms domain)
      fun syn_decls u = Aux.foldl (op @) ([], map Syn_mng.get_syn_decls domain)
      fun trrules u = Aux.foldl (op @) ([], map Syn_mng.get_trrules domain)
      fun parse_ast_trans u = Aux.foldl (op @) ([], map Syn_mng.get_parse_ast_trans domain)
      fun parse_trans u = Aux.foldl (op @) ([], map Syn_mng.get_parse_trans domain)
      fun print_trans u = Aux.foldl (op @) ([], map Syn_mng.get_print_trans domain)
      fun print_ast_trans u = Aux.foldl (op @) ([], map Syn_mng.get_print_ast_trans domain)



      fun ins_nt (thy, nt) = (if (Syn_mng.is_transl1 nt thy) then thy
                              else (let val _ = Aux.info ("... adding nonterminal: " ^ nt)
                                        val _ = Syn_mng.add_transl1 nt trgnm
                                        val _ = Syn_mng.add_nonterm (trgnm, nt)
                                    in 
                                     Sign.add_nonterminals [Binding.name nt] thy
                                    end))
      val trg2 = Aux.foldl ins_nt (target, nts ())

      fun transl_tp (Type(nm, ts)) = 
          let val unm = Aux.unpack nm
          in
           if ((hd unm) mem domain) then Type (Aux.pack (trgnm::(tl unm)), map transl_tp ts)
           else Type (nm, map transl_tp ts)    
          end
        | transl_tp var = var 

      fun transl_syn_decl (thy, (s, b, nm, tp, mxfx)) =
       (if (Syn_mng.is_transl1 nm thy) then thy
        else (let val ntp = Aux.unvarifyT (transl_tp (map_type tp))
                  val _ = Aux.info ("... adding syntax declaration: " ^ nm ^ 
                                       " :: \"" ^ (Syntax.string_of_typ_global thy ntp) ^ 
                                       "\"   " ^ (Syn_mng.string_of_mixfix mxfx))
                  val _ = Syn_mng.add_transl1 nm trgnm
                  val _ = Syn_mng.add_syn_decl (trgnm, (s, b, nm, ntp, mxfx))
              in 
               Sign.add_modesyntax_i (s, b) [(nm, ntp, mxfx)] thy
              end))

      val trg3 = Aux.foldl transl_syn_decl (trg2, syn_decls ()) 


(* ***** *)

      fun show_trrule (Syntax.ParseRule (l, r)) = ("\"" ^ snd l ^ "\"" ^ " => " ^ "\"" ^ snd r ^ "\"" )
        | show_trrule (Syntax.PrintRule (l, r)) = ("\"" ^ snd l ^ "\"" ^ " <= " ^ "\"" ^ snd r ^ "\"")
        | show_trrule (Syntax.ParsePrintRule (l, r)) = ("\"" ^ snd l ^ "\"" ^ " == " ^ "\"" ^ snd r ^ "\"")


      fun ins_trrule (thy, trrule) = 
       let val rep = show_trrule trrule
       in
        if (Syn_mng.is_transl1 rep thy) then thy
        else (let val _ = Aux.info ("... adding translation: " ^ rep)
                  val _ = Syn_mng.add_trrule (trgnm, trrule)
                  val _ = Syn_mng.add_transl1 rep trgnm
              in
                Sign.add_trrules [trrule] thy
              end)
       end
      val trg4 = Aux.foldl ins_trrule (trg3, trrules ())
 
      fun ins_parse_ast (thy, t) = 
       let val (b, (rep, pos)) = t
       in
        if (Syn_mng.is_transl1 rep thy) then thy
        else (let val _ = Aux.info ("... adding parse_ast_translation: " ^ rep)
                  val _ = Syn_mng.add_parse_ast_trans (trgnm, t)
                  val _ = Syn_mng.add_transl1 rep trgnm
              in
               IsarCmd.parse_ast_translation t thy
              end)
       end
      val trg5 = Aux.foldl ins_parse_ast (trg4, parse_ast_trans ())
     
      fun ins_parse (thy, t) = 
       let val (b, (rep, pos)) = t
       in
        if (Syn_mng.is_transl1 rep thy) then thy
        else (let val _ = Aux.info ("... adding parse_translation: " ^ rep)
                  val _ = Syn_mng.add_parse_trans (trgnm, t)
                  val _ = Syn_mng.add_transl1 rep trgnm
              in
               IsarCmd.parse_translation t thy
              end)
       end
      val trg6 = Aux.foldl ins_parse (trg5, parse_trans ())

      fun ins_print (thy, t) = 
       let val (b, (rep, pos)) = t
       in
        if (Syn_mng.is_transl1 rep thy) then thy
        else (let val _ = Aux.info ("... adding print_translation: " ^ rep)
                  val _ = Syn_mng.add_print_trans (trgnm, t)
                  val _ = Syn_mng.add_transl1 rep trgnm
              in
               IsarCmd.print_translation t thy
              end)
       end
      val trg7 = Aux.foldl ins_print (trg6, print_trans ())
   
      fun ins_print_ast (thy, t) = 
       let val (b, (rep, pos)) = t
       in
        if (Syn_mng.is_transl1 rep thy) then thy
        else (let val _ = Aux.info ("... adding print_ast_translation: " ^ rep)
                  val _ = Syn_mng.add_print_ast_trans (trgnm, t)
                  val _ = Syn_mng.add_transl1 rep trgnm
              in
               IsarCmd.print_ast_translation t thy
              end)
       end
      val trg8 = Aux.foldl ins_print_ast (trg7, print_ast_trans ())

 in
  trg8
 end

end